==> Dressing_1.0.0.202512301119-1213.jar: src/geometry/EarClipper.java <==
package geometry;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Vector;

import com.badlogic.gdx.math.Vector2;

import dressing.mathutils.Edge;
import dressing.mathutils.MathUtilities;
import dressing.mathutils.Triangle;

public class EarClipper {
	
	public static  ArrayList<Triangle> triangulate(ArrayList<Vector2> vertices){
        ArrayList<Triangle> ears = new ArrayList<Triangle>();
        ArrayList<Vector2>  verticesCpy = new ArrayList<Vector2>();
        verticesCpy.addAll(vertices);
        earClip(verticesCpy,ears);
        return ears;
    }

    public static  ArrayList<Triangle> triangulate(ArrayList<Vector2> polygonVertices, ArrayList<Vector2> holeVertices){
        ArrayList<Triangle> ears = new ArrayList<Triangle>();
        if(holeVertices == null || holeVertices.isEmpty()) {
        	 earClip(polygonVertices,ears);
        }else {
        	Vector2[] cut= findMutuallyVisibleVertices(polygonVertices,holeVertices);
            ArrayList<Vector2> vertices = new ArrayList<>();
            for(int i = 0;  i <= polygonVertices.indexOf(cut[1]); i++){
                vertices.add(polygonVertices.get(i));
            }
            int index = holeVertices.indexOf(cut[0]);
            for(int i =0; i < holeVertices.size(); i++){
                vertices.add(holeVertices.get(index));
                index++;
                index = index%holeVertices.size();
            }
            vertices.add(cut[0]);
            for(int i = polygonVertices.indexOf(cut[1]); i < polygonVertices.size(); i++){
                vertices.add(polygonVertices.get(i));
            }
            ArrayList<Vector2>  verticesCpy = new ArrayList<Vector2>();
            verticesCpy.addAll(vertices);
            earClip(verticesCpy,ears);
       } 
       return ears;
    }


    //before u call this recursive method make sure the vertices re in a positive winding ccw
    public static void earClip(ArrayList<Vector2> vertices,ArrayList<Triangle> ears) {
        if(vertices.size() > 2){
            Vector2 earVertex = null;
            for(Vector2 vertex: vertices){
                earVertex = null;
                if(isAnEar(vertex, vertices)){
                    int index = vertices.indexOf(vertex);
                    int previousIndex = (index -1)%vertices.size();
                    if(previousIndex < 0)
                        previousIndex += vertices.size();
                    Vector2 previousVertex = vertices.get(previousIndex);
                    Vector2 nextVertex = vertices.get((index + 1) % vertices.size());
                    Triangle triangle = new Triangle(vertex,previousVertex,nextVertex);
                    ears.add(triangle);
                    earVertex = vertex;
                    break;
                }
            }
            if(earVertex != null){
                vertices.remove(earVertex);
                earClip(vertices,ears);
            }
        }
    }

    public static boolean isAnEar(Vector2 vertex, ArrayList<Vector2> polygonVertices) {
    	
        int index = polygonVertices.indexOf(vertex);
        if(index >= 0){
            int previousIndex = (index -1)%polygonVertices.size();
            if(previousIndex < 0)
                previousIndex += polygonVertices.size();
            int nextIndex = (index + 1) % polygonVertices.size();
            Vector2 vcp = polygonVertices.get(previousIndex).cpy().sub(vertex);
            Vector2 vcn = polygonVertices.get(nextIndex).cpy().sub(vertex);
            
            //float angle = MathUtilities.signedAngle(vcp, vcn);
            float angle = vcn.angle(vcp);
            if(angle <= 0)
                return false;
            Triangle triangle = new Triangle(vertex, polygonVertices.get(previousIndex),polygonVertices.get(nextIndex));
            for(Vector2 vertexX: polygonVertices){
                if(vertexX == vertex  || vertexX == polygonVertices.get(previousIndex) || vertexX == polygonVertices.get(nextIndex))
                    continue;
                if(triangle.contains(vertexX))
                    return false;
            }
            return true;
        }
        return false;
    }

    public static Vector2[] findMutuallyVisibleVertices(ArrayList<Vector2> outerPoly,ArrayList<Vector2> innerPoly){
        Vector2[] mutuallyVisibleVertices = null;
        ArrayList<Edge> edges = new ArrayList<>();
        for(int i = 0; i < outerPoly.size(); i++){
            int nextIndex = (i+1)%outerPoly.size();
            edges.add(new Edge(outerPoly.get(i), outerPoly.get(nextIndex)));
        }
        for(int i = 0; i < innerPoly.size(); i++){
            int nextIndex = (i+1)%innerPoly.size();
            edges.add(new Edge(innerPoly.get(i), innerPoly.get(nextIndex)));
        }
        for(Vector2 innerVertex: innerPoly){
            boolean terminate = false;
            for(Vector2 outerVertex: outerPoly){
                Edge edge = new Edge(innerVertex, outerVertex);
                boolean mutuallyVisible = true;
                for(Edge edgeX: edges){
                    Vector2 intersectionpoint = MathUtilities.getIntersectionPoint(edge, edgeX);
                    if(edgeX.getV0().epsilonEquals(intersectionpoint,0.0001f) || edgeX.getV1().epsilonEquals(intersectionpoint,0.0001f))
                        continue;
                    if(edgeX.contains(intersectionpoint) && edge.contains(intersectionpoint)){
                        mutuallyVisible = false;
                        break;
                    }
                }
                if(mutuallyVisible){
                    mutuallyVisibleVertices = new Vector2[]{innerVertex,outerVertex};
                    terminate = true;
                    break;
                }
            }
            if(terminate)
                break;
        }
        return mutuallyVisibleVertices;
    }
    
    public static void main(String...strings) {
    	Vector2[] direct = { new Vector2(0, 0), 
    			new Vector2(100, 0),
    			new Vector2(100,100),
    			new Vector2(0, 100)
    	};
    	
    	Vector2[] indirect = { new Vector2(0, 0), 
    			new Vector2(0, 100),
    			new Vector2(100,100),
    			new Vector2(100, 0),
    	};
    	
    	
    	
    	ArrayList<Vector2> listDirect = new ArrayList<Vector2>(Arrays.asList(direct));
    	MathUtilities.setWinding(listDirect, 1);
    	
    	ArrayList<Vector2> listinDirect = new ArrayList<Vector2>(Arrays.asList(indirect));
    	MathUtilities.setWinding(listinDirect, 1);
    	
    	ArrayList<Triangle> ears = new ArrayList<Triangle>();
    	EarClipper.earClip(listDirect, ears);
    	EarClipper.earClip(listinDirect, ears);

    	for(Triangle triangle: ears)
    		System.out.println(triangle);
    	
    	


    }

}

==> Dressing_1.0.0.202601161112-1217.jar: src/geometry/EarClipper.java <==
package geometry;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Vector;

import com.badlogic.gdx.math.Vector2;

import dressing.mathutils.Edge;
import dressing.mathutils.MathUtilities;
import dressing.mathutils.Triangle;

public class EarClipper {
	
	public static  ArrayList<Triangle> triangulate(ArrayList<Vector2> vertices){
        ArrayList<Triangle> ears = new ArrayList<Triangle>();
        ArrayList<Vector2>  verticesCpy = new ArrayList<Vector2>();
        verticesCpy.addAll(vertices);
        earClip(verticesCpy,ears);
        return ears;
    }

    public static  ArrayList<Triangle> triangulate(ArrayList<Vector2> polygonVertices, ArrayList<Vector2> holeVertices){
        ArrayList<Triangle> ears = new ArrayList<Triangle>();
        if(holeVertices == null || holeVertices.isEmpty()) {
        	 earClip(polygonVertices,ears);
        }else {
        	Vector2[] cut= findMutuallyVisibleVertices(polygonVertices,holeVertices);
            ArrayList<Vector2> vertices = new ArrayList<>();
            for(int i = 0;  i <= polygonVertices.indexOf(cut[1]); i++){
                vertices.add(polygonVertices.get(i));
            }
            int index = holeVertices.indexOf(cut[0]);
            for(int i =0; i < holeVertices.size(); i++){
                vertices.add(holeVertices.get(index));
                index++;
                index = index%holeVertices.size();
            }
            vertices.add(cut[0]);
            for(int i = polygonVertices.indexOf(cut[1]); i < polygonVertices.size(); i++){
                vertices.add(polygonVertices.get(i));
            }
            ArrayList<Vector2>  verticesCpy = new ArrayList<Vector2>();
            verticesCpy.addAll(vertices);
            earClip(verticesCpy,ears);
       } 
       return ears;
    }


    //before u call this recursive method make sure the vertices re in a positive winding ccw
    public static void earClip(ArrayList<Vector2> vertices,ArrayList<Triangle> ears) {
        if(vertices.size() > 2){
            Vector2 earVertex = null;
            for(Vector2 vertex: vertices){
                earVertex = null;
                if(isAnEar(vertex, vertices)){
                    int index = vertices.indexOf(vertex);
                    int previousIndex = (index -1)%vertices.size();
                    if(previousIndex < 0)
                        previousIndex += vertices.size();
                    Vector2 previousVertex = vertices.get(previousIndex);
                    Vector2 nextVertex = vertices.get((index + 1) % vertices.size());
                    Triangle triangle = new Triangle(vertex,previousVertex,nextVertex);
                    ears.add(triangle);
                    earVertex = vertex;
                    break;
                }
            }
            if(earVertex != null){
                vertices.remove(earVertex);
                earClip(vertices,ears);
            }
        }
    }

    public static boolean isAnEar(Vector2 vertex, ArrayList<Vector2> polygonVertices) {
    	
        int index = polygonVertices.indexOf(vertex);
        if(index >= 0){
            int previousIndex = (index -1)%polygonVertices.size();
            if(previousIndex < 0)
                previousIndex += polygonVertices.size();
            int nextIndex = (index + 1) % polygonVertices.size();
            Vector2 vcp = polygonVertices.get(previousIndex).cpy().sub(vertex);
            Vector2 vcn = polygonVertices.get(nextIndex).cpy().sub(vertex);
            
            //float angle = MathUtilities.signedAngle(vcp, vcn);
            float angle = vcn.angle(vcp);
            if(angle <= 0)
                return false;
            Triangle triangle = new Triangle(vertex, polygonVertices.get(previousIndex),polygonVertices.get(nextIndex));
            for(Vector2 vertexX: polygonVertices){
                if(vertexX == vertex  || vertexX == polygonVertices.get(previousIndex) || vertexX == polygonVertices.get(nextIndex))
                    continue;
                if(triangle.contains(vertexX))
                    return false;
            }
            return true;
        }
        return false;
    }

    public static Vector2[] findMutuallyVisibleVertices(ArrayList<Vector2> outerPoly,ArrayList<Vector2> innerPoly){
        Vector2[] mutuallyVisibleVertices = null;
        ArrayList<Edge> edges = new ArrayList<>();
        for(int i = 0; i < outerPoly.size(); i++){
            int nextIndex = (i+1)%outerPoly.size();
            edges.add(new Edge(outerPoly.get(i), outerPoly.get(nextIndex)));
        }
        for(int i = 0; i < innerPoly.size(); i++){
            int nextIndex = (i+1)%innerPoly.size();
            edges.add(new Edge(innerPoly.get(i), innerPoly.get(nextIndex)));
        }
        for(Vector2 innerVertex: innerPoly){
            boolean terminate = false;
            for(Vector2 outerVertex: outerPoly){
                Edge edge = new Edge(innerVertex, outerVertex);
                boolean mutuallyVisible = true;
                for(Edge edgeX: edges){
                    Vector2 intersectionpoint = MathUtilities.getIntersectionPoint(edge, edgeX);
                    if(edgeX.getV0().epsilonEquals(intersectionpoint,0.0001f) || edgeX.getV1().epsilonEquals(intersectionpoint,0.0001f))
                        continue;
                    if(edgeX.contains(intersectionpoint, 1e-3f) && edge.contains(intersectionpoint, 1e-3f)){
                        mutuallyVisible = false;
                        break;
                    }
                }
                if(mutuallyVisible){
                    mutuallyVisibleVertices = new Vector2[]{innerVertex,outerVertex};
                    terminate = true;
                    break;
                }
            }
            if(terminate)
                break;
        }
        return mutuallyVisibleVertices;
    }
    
    public static void main(String...strings) {
    	Vector2[] direct = { new Vector2(0, 0), 
    			new Vector2(100, 0),
    			new Vector2(100,100),
    			new Vector2(0, 100)
    	};
    	
    	Vector2[] indirect = { new Vector2(0, 0), 
    			new Vector2(0, 100),
    			new Vector2(100,100),
    			new Vector2(100, 0),
    	};
    	
    	
    	
    	ArrayList<Vector2> listDirect = new ArrayList<Vector2>(Arrays.asList(direct));
    	MathUtilities.setWinding(listDirect, 1);
    	
    	ArrayList<Vector2> listinDirect = new ArrayList<Vector2>(Arrays.asList(indirect));
    	MathUtilities.setWinding(listinDirect, 1);
    	
    	ArrayList<Triangle> ears = new ArrayList<Triangle>();
    	EarClipper.earClip(listDirect, ears);
    	EarClipper.earClip(listinDirect, ears);

    	for(Triangle triangle: ears)
    		System.out.println(triangle);
    	
    	


    }

}
